"""
Database-guided multi-robot motion planning

This module is essentially one big path tracking algorithm. 
It uses MPC to path track each of the robots, while looking ahead to 
identify and resolve conflicts.
"""
import time as t
import pygame
from shapely.geometry import Polygon, Point

from guided_mrmp.utils import Env
from guided_mrmp.utils.helpers import *

from guided_mrmp.conflict_resolvers import TrajOptResolver,TrajOptDBResolver

from guided_mrmp.controllers.utils import get_ref_trajectory, compute_path_from_wp

from guided_mrmp.controllers.mpc import MPC

from guided_mrmp.utils.library import Library

def initialize_libraries(library_fnames=["guided_mrmp/database/2x3_library","guided_mrmp/database/3x3_library","guided_mrmp/database/5x2_library"]):
    """
    Load the 2x3, 3x3, and 2x5 libraries. Store them in self.lib-x- 
    Inputs: 
        library_fnames - the folder containing the library files
    """
    # Create each of the libraries
    print(f"Loading libraries. This usually takes about 10 seconds...")
    lib_2x3 = Library(library_fnames[0])
    lib_2x3.read_library_from_file()
    
    lib_3x3 = Library(library_fnames[1])
    lib_3x3.read_library_from_file()

    
    lib_2x5 = Library(library_fnames[2])
    lib_2x5.read_library_from_file()

    return [lib_2x3, lib_3x3, lib_2x5]



class GuidedMRMP:
    def __init__(self, env, robots, dynamics_models, T, DT, settings):
        """
        inputs:
            - robots (list): list of Robot class objects
            - env (Env): the environment
        """
        self.robots = robots
        self.dynamics_models = dynamics_models
        self.env = env
        self.guide_paths = [[]]*len(robots)
        self.settings = settings

        self.T = T
        self.DT = DT
        self.K = int(T / DT)

        # for MPC
        self.Q = self.settings['model_predictive_controller']['Q']  # state error cost
        self.Qf = self.settings['model_predictive_controller']['Qf']  # state final error cost
        self.R = self.settings['model_predictive_controller']['R']  # input cost
        self.P = self.settings['model_predictive_controller']['P']   # input rate of change cost

    
        for idx,r in enumerate(self.robots):
            xs = []
            ys = []
            for node in r.rrtpath:
                xs.append(node[0])
                ys.append(node[1])

            waypoints = [xs,ys]

            self.guide_paths[idx] = compute_path_from_wp(waypoints[0], waypoints[1], .05)

        self.scaling_factor = self.settings['simulator']['scaling_factor']

        self.libs = initialize_libraries()

    def ego_to_global(self, robot, mpc_out):
        """
        transforms optimized trajectory XY points from ego (robot) reference
        into global (map) frame

        Args:
            mpc_out ():
        """
        # Extract x, y, and theta from the state
        x = robot.current_position[0]
        y = robot.current_position[1]
        theta = robot.current_position[2]

        # Rotation matrix to transform points from ego frame to global frame
        Rotm = np.array([
            [np.cos(theta), -np.sin(theta)],
            [np.sin(theta), np.cos(theta)]
        ])

        # Initialize the trajectory array (only considering XY points)
        trajectory = mpc_out[0:2, :]

        # Apply rotation to the trajectory points
        trajectory = Rotm.dot(trajectory)

        # Translate the points to the robot's position in the global frame
        trajectory[0, :] += x
        trajectory[1, :] += y

        return trajectory
    
    def find_all_conflicts(self, desired_trajs, dt):
        """
        Check for any conflicts that will occur between robots in the next time horizon
        """
        conflicts = []

        for r1_idx, r1 in enumerate(self.robots):
            # control = desired_controls[r1_idx]
            traj1 = desired_trajs[r1_idx]
            traj1 = list(zip(traj1[0],traj1[1]))
            
            for r2_idx, r2 in enumerate(self.robots[r1_idx+1:]):
                # control = desired_controls[r2_idx]
                traj2 = desired_trajs[r2_idx+r1_idx+1]
                traj2 = list(zip(traj2[0],traj2[1]))

                for p1, p2 in zip(traj1, traj2):
                    circ1 = Point(p1[0],p1[1])
                    circ1 = circ1.buffer(r1.radius)
                    circ2 = Point(p2[0],p2[1])
                    circ2 = circ2.buffer(r2.radius)
                    # check if the robots overlap
                    if circ1.intersects(circ2): 
                        if (r1,r2) not in conflicts and (r2,r1) not in conflicts:
                            # print(f"conflict between {r1.label} and {r2.label}")
                            conflicts.append((r1,r2))
                
                

        return conflicts

    def get_next_controls_and_trajectories_parallel(self,screen):
        """
        Get the next control for each robot.

        """
        from joblib import Parallel, delayed
        
        next_controls = []
        next_trajs = []

        def process_robot(idx, r):
            state = r.current_position
            path = self.guide_paths[idx]

            # Get Reference_traj -> inputs are in worldframe
            target_traj = get_ref_trajectory(np.array(state), 
                                            np.array(path), 
                                            r.target_v, 
                                            self.T, 
                                            self.DT)
            
            mpc = MPC(self.dynamics_models[idx], self.T, self.DT, self.Q, self.Qf, self.R, self.P, self.settings['model_predictive_controller'])

            if r.last_mpc_trajectory is not None:
                # dynamics w.r.t robot frame
                initial_guess = {'X': r.last_mpc_trajectory, 'U': r.last_mpc_controls}
                curr_state = np.array([0, 0, 0])
                x_mpc, u_mpc = mpc.step(curr_state, target_traj, np.array(r.control), initial_guess)

            else:
                # dynamics w.r.t robot frame
                curr_state = np.array([0, 0, 0])
                x_mpc, u_mpc = mpc.step(curr_state, target_traj, np.array(r.control))

            # only the first one is used to advance the simulation
            control = [u_mpc[0, :], u_mpc[1, :]]

            r.last_mpc_trajectory = x_mpc
            r.last_mpc_controls = u_mpc
            
            return np.asarray(control), self.ego_to_global(r, x_mpc)

        with Parallel(n_jobs=12) as parallel:
            results = parallel(delayed(process_robot)(idx, r) for idx, r in enumerate(self.robots))

        next_controls = [r[0] for r in results]
        next_trajs = [r[1] for r in results]


        return next_controls, next_trajs
    
    def get_next_controls_and_trajectories(self,screen):
        """
        Get the next control for each robot.
        """

        next_controls = []
        next_trajs = []
        for idx, r in enumerate(self.robots):

            state = r.current_position
            path = self.guide_paths[idx]

            # Get Reference_traj -> inputs are in worldframe
            target_traj = get_ref_trajectory(np.array(state), 
                                             np.array(path), 
                                             r.target_v, 
                                             self.T, 
                                             self.DT)
            
            mpc = MPC(self.dynamics_models[idx], self.T, self.DT, self.Q, self.Qf, self.R, self.P, self.settings['model_predictive_controller'])

            # dynamics w.r.t robot frame
            curr_state = np.array([0, 0, 0])
            x_mpc, u_mpc = mpc.step(curr_state, target_traj, np.array(r.control))

            self.add_vis_target_traj(screen, self.scaling_factor, r, x_mpc)
            
            # only the first one is used to advance the simulation
            control = [u_mpc[0, :], u_mpc[1, :]]
            
            next_controls.append(np.asarray(control))

            # print(f"control = {u_mpc}")
            next_trajs.append(self.ego_to_global(r, x_mpc))
        return next_controls, next_trajs

    def advance(self, screen, state, time, dt=0.1):
        """
        Advance the simulation by one timestep.

        inputs:
            - state (list): list of tuples, where each tuple is the state of a robot
            - time (float): current time
            - t (float): time step
        """

        # get the next control for each robot
        next_desired_controls, trajectories = self.get_next_controls_and_trajectories(screen)

        # find all the conflicts at the next time horizon
        conflicts = self.find_all_conflicts(trajectories, dt)
        conflicts = []

        if len(conflicts) == 0:
            # no conflicts, return the desired controls
            controls = []

            for control in next_desired_controls:
                next_controls = [control[0][0], control[1][0]]
                controls.append(np.asarray(next_controls))

            for r, next_control in zip(self.robots, controls):
                r.control = next_control
            return False, controls

        # resolve the conflicts using the database
        # resolver = TrajOptDBResolver(10, 60, conflicts, self.robots)
        rad = self.robots[0].radius
        resolver = TrajOptDBResolver(rad*2, 5, conflicts, self.robots, trajectories, dt, rad,self.env,1,1,1,self.libs[0],self.libs[1],self.libs[2]) 
        next_desired_controls, trajs = resolver.get_local_controls(next_desired_controls)

        self.add_vis_conflict_free_traj(screen, self.scaling_factor, trajs)

        # data = {
        #     "conflicts": conflicts,
        #     "dt": dt,
        #     "env": self.env,
        #     "next_desired_controls": next_desired_controls,
        #     "robots": self.robots,
        #     "trajectories": trajs
        # }

        # import yaml
        # with open("tests/db_opt_data1.yaml", "w") as file:
        #     yaml.dump(data, file)
            

        # return the valid controls
        for r, next_control in zip(self.robots, next_desired_controls):
            r.control = [next_control[0][-1], next_control[1][-1]]
        
        return True, next_desired_controls


    def add_vis_guide_paths(self, screen, scaling_factor):
        """
        Add the visualization to the screen.
        """
        
        for r in self.robots:
            path = self.guide_paths[r.label]
            xs = path[0]
            ys = path[1]
            # for node in path:
            #     pygame.draw.circle(screen, r.color, (int(node[0]), int(node[1])), 2)
            for i in range(len(xs)-1):
                pygame.draw.line(screen, r.color, (xs[i]*scaling_factor,ys[i]*scaling_factor), 
                                 (xs[i+1]*scaling_factor,ys[i+1]*scaling_factor), 2)

    def add_vis_grid(self, screen, grid_size, top_left, cell_size):
        """
        Add the visualization to the screen.
        """
         # draw the grid
        line_len = grid_size*cell_size
        for i in range(grid_size+1):
            xs = [top_left[0] + i*cell_size,top_left[0] + i*cell_size]
            ys = [top_left[1], top_left[1]-line_len]
            pygame.draw.line(screen, (0,0,0), (xs[0],ys[0]), (xs[1],ys[1]), 2)

            xs = [top_left[0], top_left[0]+line_len]
            ys = [top_left[1] - i*cell_size,top_left[1] - i*cell_size]
            
            pygame.draw.line(screen, (0,0,0), (xs[0],ys[0]), (xs[1],ys[1]), 2)

    def add_vis_target_traj(self,screen, scaling_factor, robot, traj):
        """
        Add the visualization to the screen.
        """
        traj = self.ego_to_global(robot, traj)
        for i in range(len(traj[0])-1):
            x = int(traj[0,i])
            y = int(traj[1,i])
            next_x = int(traj[0,i+1])
            next_y = int(traj[1,i+1])
            pygame.draw.line(screen, (255,0,0), (x*scaling_factor,y*scaling_factor), 
                             (next_x*scaling_factor,next_y*scaling_factor), 2)

    def add_vis_conflict_free_traj(self,screen, scaling_factor, trajs):
        """
        Add the visualization to the screen.
        """
        # traj = self.ego_to_global(robot, traj)
        for traj in trajs:
            # print(traj)
            for i in range(len(traj[0])-1):
                x = int(traj[0][i])
                y = int(traj[1][i])
                next_x = int(traj[0][i+1])
                next_y = int(traj[1][i+1])
                pygame.draw.line(screen, (0,255,0), (x*scaling_factor,y*scaling_factor), 
                                (next_x*scaling_factor,next_y*scaling_factor), 2)


if __name__ == "__main__":

    # create the environment
    env = Env()

    # starts, goals = create_random_starts_and_goals(env, 4)
    starts = [[0,0.1,0],[10,5.1,0]]
    goals = [[10,3,0],[5,1,0]]

    alg = GuidedMRMP(Env)
    alg.run(starts, goals)